#! /usr/bin/python3
# -*- coding: utf-8 -*-

# Copyright © 2015 Collabora Ltd.
#
# SPDX-License-Identifier: MPL-2.0
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import unittest
import os
import hashlib
import sys

from gi.repository import GLib
from gi.repository import Gio
from gi.repository import Grl

# import from parent directory
sys.path.insert(0, os.path.join(os.path.dirname(__file__), os.pardir))
from common import TrackerIndexer
from common import LONG_JPEG_NAME


class TrackerTest(unittest.TestCase):
    def setUp(self):
        self.loop = GLib.MainLoop.new(None, False)
        self.homedir = os.path.expanduser("~")
        self.indexer = TrackerIndexer()
        self.indexer.copy_medias(self.homedir)

        # Monitor thumbnail creation to know when it's done. The DBus API
        # doesn't have a method to query the initial state but we can be
        # reasonably sure it's not thumbnailing anything at this point.
        self.tumbler_queue = []
        self.tumbler = Gio.DBusProxy.new_for_bus_sync(
            Gio.BusType.SESSION,
            Gio.DBusProxyFlags.DO_NOT_LOAD_PROPERTIES,
            None,
            'org.freedesktop.thumbnails.Thumbnailer1',
            '/org/freedesktop/thumbnails/Thumbnailer1',
            'org.freedesktop.thumbnails.Thumbnailer1',
            None)
        self.tumbler.connect('g-signal', self.tumbler_signal_cb)

        # Spy on unicast signals that weren't meant for us, because
        # the Thumbnailer API uses those, and we want to use them to
        # determine when it has finished. GDBus doesn't have
        # high-level API for this sort of nonsense so we do it the hard way.
        match_rule = ("type=signal,"
                      "sender='org.freedesktop.thumbnails.Thumbnailer1',"
                      "eavesdrop=true")
        conn = self.tumbler.get_connection()
        conn.call_sync('org.freedesktop.DBus',
                       '/org/freedesktop/DBus',
                       'org.freedesktop.DBus',
                       'AddMatch',
                       GLib.Variant('(s)', (match_rule,)),
                       None, Gio.DBusCallFlags.NONE, -1, None)

    def tearDown(self):
        # Keep everything in place for further manual checks
        pass

    def tumbler_signal_cb(self, proxy, sender_name, signal_name, parameters):
        print('TrackerTest: Received tumbler signal:', signal_name)

        if signal_name == 'Started':
            child = parameters.get_child_value(0)
            self.tumbler_queue.append(child.get_uint32())
        elif signal_name == 'Finished':
            child = parameters.get_child_value(0)
            self.tumbler_queue.remove(child.get_uint32())
        elif signal_name == 'Error':
            child = parameters.get_child_value(1)
            uris = child.get_strv()
            child = parameters.get_child_value(3)
            msg = child.get_string()
            raise Exception("Error creating thumbnail for %s: %s" %
                            ', '.join(uris), msg)

    def tumbler_get_supported_cb(self, source, result):
        self.tumbler.call_finish(result)
        self.loop.quit()

    def tracker_config_tests(self):
        print("TrackerTest: config tests")
        settings = Gio.Settings.new("org.freedesktop.Tracker.Miner.Files")
        self.assertEqual(settings.get_boolean('index-removable-devices'), True)
        self.assertEqual(settings.get_int('removable-days-threshold'), 60)
        self.assertEqual(settings.get_int('initial-sleep'), 0)
        self.assertEqual(settings.get_string('sched-idle'), 'never')
        self.assertEqual(settings.get_boolean('enable-monitors'), True)
        # self.assertEqual(settings.get_boolean('enable-writeback'), False)

    def tracker_journal_tests(self):
        print("TrackerTest: journal tests")
        path = ('%s/.local/share/tracker/data/tracker-store.journal' %
                self.homedir)
        self.assertFalse(os.path.isfile(path))

    def tracker_initial_tests(self):
        print("TrackerTest: initial tests")
        self.indexer.assert_all_indexed(self.homedir)

    def tracker_update_tests(self):
        print("TrackerTest: update tests")

        # Create a new file and assert it gets indexed
        filename = self.homedir + '/Documents/something.txt'
        with open(filename, 'w') as f:
            f.write('something')
        self.indexer.wait(True)
        self.indexer.assert_indexed(
            filename, '?urn nie:plainTextContent "something"')

        # Modify the file should re-index it
        with open(filename, 'w') as f:
            f.write('something else')
        self.indexer.wait(True)
        self.indexer.assert_indexed(
            filename, '?urn nie:plainTextContent "something else"')

        # Delete file and assert it's not indexed anymore
        os.remove(filename)
        self.indexer.wait(False)
        self.indexer.assert_not_indexed(filename)

    def assert_has_thumbnail(self, filename):
        # Note that this is the path for local storage only, not for removable
        # devices.
        uri = 'file://%s/%s' % (self.homedir, filename)
        uri_hash = hashlib.md5(bytes(uri, encoding='UTF-8')).hexdigest()
        path = '%s/.cache/thumbnails/normal/%s.png' % (self.homedir, uri_hash)
        self.assertTrue(os.path.isfile(path))

    def thumbnail_tests(self):
        print("TrackerTest: thumbnail tests")

        # Drain the DBus queue to make sure we received all signals
        self.tumbler.call('GetSupported',
                          GLib.Variant.new_tuple(),
                          Gio.DBusCallFlags.NONE,
                          -1,
                          None,
                          self.tumbler_get_supported_cb)
        self.loop.run()

        context = self.loop.get_context()
        while len(self.tumbler_queue) > 0:
            context.iteration(True)

        self.assert_has_thumbnail('Documents/lorem_presentation.odp')
        self.assert_has_thumbnail('Documents/lorem_spreadsheet.ods')
        self.assert_has_thumbnail('Documents/more_lorem_ipsum.odt')
        self.assert_has_thumbnail('Pictures/' + LONG_JPEG_NAME)
        self.assert_has_thumbnail('Pictures/collabora-logo-big.png')
        self.assert_has_thumbnail('Videos/big_buck_bunny_smaller.ogv')

    def grl_browse_source(self, source, media):
        ops = source.supported_operations()
        self.assertTrue((ops & Grl.SupportedOps.BROWSE) != 0)

        caps = source.get_caps(Grl.SupportedOps.BROWSE)
        options = Grl.OperationOptions.new(caps)
        options.set_count(10)
        options.set_resolution_flags(Grl.ResolutionFlags.IDLE_RELAY)

        keys = [Grl.METADATA_KEY_TITLE,
                Grl.METADATA_KEY_ALBUM,
                Grl.METADATA_KEY_ARTIST,
                Grl.METADATA_KEY_THUMBNAIL,
                Grl.METADATA_KEY_URL,
                Grl.METADATA_KEY_CHILDCOUNT]

        medias = source.browse_sync(media, keys, options)

        for media in medias:
            if isinstance(media, Grl.MediaBox):
                self.grl_browse_source(source, media)
                continue

            url = media.get_url()
            prefix = 'file://' + self.homedir + '/'
            if not url.startswith(prefix):
                continue

            filename = url[len(prefix):]
            if filename in ['Music/generic.oga',
                            'Music/generic.mp3',
                            'Music/generic.flac',
                            'Music/generic-no-artwork.mp3']:
                self.assertEqual(media.get_title(), 'GNOME Generic Sound')
                self.assertEqual(media.get_artist(), 'Conrad Parker')
                self.assertEqual(media.get_album(), 'GNOME Audio')
            elif filename in ['Music/generic.wav',
                              'Pictures/' + LONG_JPEG_NAME,
                              'Pictures/collabora-logo-big.png',
                              'Videos/big_buck_bunny_smaller.ogv']:
                title = filename.split('/')[1]
                self.assertEqual(media.get_title(), title)

    def grl_source_added_cb(self, registry, source):
        self.grl_browse_source(source, None)
        self.loop.quit()

    def tracker_grilo_tests(self):
        print("TrackerTest: grilo tests")
        registry = Grl.Registry.get_default()
        registry.load_plugin_by_id('grl-tracker')
        registry.connect('source-added', self.grl_source_added_cb)
        self.loop.run()

    # This is the only test case to make only one setup of tracker
    def test_all(self):
        self.tracker_config_tests()
        self.tracker_journal_tests()
        self.indexer.start()
        self.tracker_initial_tests()
        self.tracker_update_tests()
        self.tracker_grilo_tests()
        self.thumbnail_tests()

if __name__ == "__main__":
    Grl.init([])
    unittest.main()
